# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
input validation module:
This module validates user input, it throwes exceptions if validation fails
"""

from SentinelExceptions import InputError

# pylint: disable-msg=R0903
class InputValidation():
    """ This class validates user inputs """
    @staticmethod

    def validate_input(name, text):
        """ Validating input, no None or empty """
        if not text:
            raise InputError(name)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Config Reader:
This module is used to read JSON config file populated by Azure Notebooks API.
"""

import json

# pylint: disable-msg=R0903
class ConfigReader:
    """ class to read configuration """

    # pylint: disable-msg=E0213
    def read_config_values(file_path):
        """ read configuration """
        with open(file_path) as json_file:
            if json_file:
                json_config = json.load(json_file)
                return (json_config["tenant_id"],
                        json_config["subscription_id"],
                        json_config["resource_group"],
                        json_config["workspace_id"],
                        json_config["workspace_name"])
        return None



# pylint: disable-msg=C0103
"""
SentinelUtils: This package provides utility methods in general
"""

# __init__.py
from .config_reader import ConfigReader
from .version_management import VersionInformation, ModuleVersionCheck
from .obfuscation_utility import ObfuscationUtility
from .input_validation import InputValidation



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Azure LogAnalytics Helper:
This module provides helper methods to initialize and 
manipulate LogAnalyticsManagementClient object.
Workspace is the focal point.
"""


class LogAnalyticsHelper():
    """ Helper class for Log Analytics """

    def __init__(self, la_client):
        self.la_client = la_client

    def get_workspace_name_list(self):
        """ retrieve L.A. workspace names as a list """
        return sorted([ws.name for ws in self.la_client.workspaces.list()])

    def get_workspace_id(self, workspace_name):
        """ retrieve L.A. workspace id based on workspace name """
        workspace = next(ws for ws in self.la_client.workspaces.list() if ws.name == workspace_name)
        return workspace.customer_id
# end of the class



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
log module:
This module provides log functionalities through Azure Application Insights
"""


from applicationinsights import TelemetryClient
from SentinelUtils.obfuscation_utility import ObfuscationUtility


class Log:
    """ The class performs log action """

    def __init__(self):
        self.seed = b'IVnGT69i43R1i6qokpVxIx_tE2MyBlMebu4yJfJh1Ow='
        # pylint: disable=line-too-long
        self.ai_code = b'gAAAAABdN3UmwZJHHxTybj0KRbKNehSo55DZ5Bi2QtohQsrEgy-WZNWDQAETVnaeJbQ4S0ltZLkebi_hhueaxl_uxYE5HheuB0ZFobq1IzgE163jUsjSLqilcqzy_uLKkSTYHAxYNLxL'

    def log(self, telemetry_instance):
        """ Log Telemetry instance """

        obfuscate = ObfuscationUtility(self.seed)
        client = TelemetryClient(obfuscate.deobfuscate_text(self.ai_code))
        client.track_trace(str(telemetry_instance))
        client.flush()

    def log_event(self, message):
        """ Log simple message in strinbg """

        obfuscate = ObfuscationUtility(self.seed)
        client = TelemetryClient(obfuscate.deobfuscate_text(self.ai_code))
        client.track_event(message)
        client.flush()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Anomaly Lookup View Helper:
This module provides helper methods for UI components.
"""

from ipywidgets import Layout, ButtonStyle, IntProgress

class AnomalyLookupViewHelper():
    """ UI Helper class for Anomaly lookup """

    @staticmethod
    def define_int_progress_bar():
        """ define progress bar """
        # pylint: disable=line-too-long
        return IntProgress(value=0, min=0, max=10, step=1, description='Loading:', bar_style='success', orientation='horizontal', position='top')

    @staticmethod
    def define_button_style():
        """ define button style """
        return ButtonStyle(button_color='#FFF', font_color='blue')

    @staticmethod
    def define_button_layout():
        """ define button layout """
        return Layout(width='auto', height='27px', border='2px solid black')



# pylint: disable-msg=C0103
# pylint: disable=line-too-long
"""
SentinelAzure: This package is developed for initializing and manipulating Python client objects of Azure Resource Management
"""

# __init__.py
from .azure_loganalytics_helper import *



# pylint: disable-msg=C0103
"""
SentinelAnomalyLookup: This package is developed for Azure Sentinel Anomaly lookup
"""

# __init__.py
from .anomaly_lookup_view_helper import AnomalyLookupViewHelper
from .anomaly_finder import AnomalyQueries, AnomalyFinder



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Hunting Bookmark Helper:
This module provides helper methods to initialize and
manipulate Hunting Bookmark (create, read, and delete).
"""
# pylint: disable-msg=C0103
# pylint: disable-msg=E0401
# pylint: disable-msg=E0602
# pylint: disable-msg=R0902
# pylint: disable-msg=R0903
# pylint: disable-msg=R0913
# pylint: disable-msg=W0201
# pylint: disable=line-too-long

import uuid
import requests
import jsons
from SentinelUtils import InputValidation


class Constants():
    """ This class holds constants """
    TYPE = 'Microsoft.SecurityInsights/Bookmarks'
    ETAG = '*'
    BOOKMARK_RESOURCE_BASE = '/subscriptions/{0}/resourceGroups/{1}/providers/Microsoft.OperationalInsights/workspaces/{2}/providers/Microsoft.SecurityInsights/bookmarks'
    BOOKMARK_ID = '/{0}'
    BOOKMARK_RESOURCE_ID = BOOKMARK_RESOURCE_BASE + BOOKMARK_ID
    ENTITY_MAPPING = '__entityMapping'
    ENTITY_TYPE = ['Bookmark'
                   'SecurityAlert',
                   'Account',
                   'Host',
                   'Malware',
                   'File',
                   'Process',
                   'CloudApplication',
                   'DnsResolution',
                   'AzureResource',
                   'FileHash',
                   'RegistryKey',
                   'RegistryValue',
                   'SecurityGroup',
                   'Url']


class BookmarkProperties():
    """
    This class holds properties for Hunting bookmark.
    Special Note:  For query_result_dict, actual value is the key, and the entity type is the value.
    """

    def __init__(self,
                 display_name,
                 query,
                 query_result_dict=None,
                 tag_list=None,
                 notes=None,
                 event_time=None,
                 query_start_time=None,
                 query_end_time=None):
        self.displayName = display_name
        self.labels = tag_list
        self.query = query
        self.queryResult = query_result_dict
        self.notes = notes
        self.eventTime = event_time
        self.queryStartTime = query_start_time
        self.queryEndTime = query_end_time
        self.bookmarkId = str(uuid.uuid4())

    @property
    def bookmarkId(self):
        """ bookmarkId getter """
        return self._bookmarkId

    @bookmarkId.setter
    def bookmarkId(self, bookmark_id):
        """ bookmarkId setter """
        self._bookmarkId = bookmark_id

    @property
    def displayName(self):
        """ displayName getter """
        return self._displayName

    @displayName.setter
    def displayName(self, display_name):
        """ displayName setter """
        InputValidation.validate_input('display_name', display_name)
        self._displayName = display_name

    @property
    def labels(self):
        """ labels getter """
        return self._labels

    @labels.setter
    def labels(self, tag_list):
        """ labels setter """
        if tag_list:
            self._labels = tag_list
        else:
            self._labels = []

    @property
    def query(self):
        """ query """
        return self._query

    @query.setter
    def query(self, query):
        """ query setter """
        InputValidation.validate_input('query', query)
        self._query = query

    @property
    def queryResult(self):
        """ queryResult """
        return self._queryResult

    @queryResult.setter
    def queryResult(self, query_result):
        """ queryResult setter """
        result_str = None
        if query_result and isinstance(query_result, dict):
            if all(elem in Constants.ENTITY_TYPE for elem in list(query_result.values())):
                query_result_dict = {}
                query_result_dict.update({Constants.ENTITY_MAPPING: query_result})
                result_str = jsons.dumps(query_result_dict)
                self._queryResult = result_str
        if not result_str:
            self._queryResult = '{}'

    @property
    def notes(self):
        """ notes getter """
        return self._notes

    @notes.setter
    def notes(self, notes):
        """ notes setter """
        self._notes = notes

    @property
    def eventTime(self):
        """ eventTime """
        return self._eventTime

    @eventTime.setter
    def eventTime(self, event_time):
        """ eventTime setter """
        self._eventTime = event_time

    @property
    def queryStartTime(self):
        """ queryStartTime """
        return self._queryStartTime

    @queryStartTime.setter
    def queryStartTime(self, query_start_time):
        """ queryStartTime setter """
        self._queryStartTime = query_start_time

    @property
    def queryEndTime(self):
        """ queryEndTime """
        return self._queryEndTime

    @queryEndTime.setter
    def queryEndTime(self, query_end_time):
        """ queryEndTime setter """
        self._queryEndTime = query_end_time


class BookmarkModel():
    """ This class holds data model for Bookmark """

    def __init__(self,
                 bookmark_name,
                 subscription_id,
                 resource_group_name,
                 workspace_name,
                 bookmark_properties):
        self.name = bookmark_name
        self.type = Constants.TYPE
        self.etag = Constants.ETAG
        self.properties = bookmark_properties
        self.bookmark_resource_base = Constants.BOOKMARK_RESOURCE_BASE.format(subscription_id, resource_group_name, workspace_name)
        self.id = self.bookmark_resource_base + Constants.BOOKMARK_ID.format(bookmark_properties.bookmarkId)

    @property
    def id(self):
        """ id getter """
        return self._id

    @id.setter
    def id(self, bookmark_resource_id):
        """ id setter """
        self._id = bookmark_resource_id

    @property
    def name(self):
        """ name getter """
        return self._name

    @name.setter
    def name(self, bookmark_name):
        """ name setter """
        InputValidation.validate_input('bookmark_name', bookmark_name)
        self._name = bookmark_name

    @property
    def type(self):
        """ type getter """
        return self._type

    @type.setter
    def type(self, t):
        """ type setter """
        self._type = t

    @property
    def etag(self):
        """ etag getter """
        return self._etag

    @etag.setter
    def etag(self, etag):
        """ etag setter """
        self._etag = etag

    @property
    def properties(self):
        """ properties getter """
        return self._properties

    @properties.setter
    def properties(self, bookmark_properties):
        """ properties setter """
        InputValidation.validate_input('bookmark_properties', bookmark_properties)
        self._properties = bookmark_properties

# pylint: disable-msg=W0703
class BookmarkHelper:
    """ This class provides CRUD methods for bookmark """

    BOOKMARK_BASE_URL = 'https://management.azure.com'
    BOOKMARK_API_VERSION = '?api-version=2019-01-01-preview'

    def __init__(self, access_token):
        self.access_token = access_token

    def _set_header(self):
        return {'Authorization': 'Bearer ' + self.access_token}

    def _set_rp_put_url(self, bookmark_model):
        return self.BOOKMARK_BASE_URL + bookmark_model.id + self.BOOKMARK_API_VERSION

    def _set_rp_get_url(self, bookmark_model):
        return self.BOOKMARK_BASE_URL + bookmark_model.bookmark_resource_base + self.BOOKMARK_API_VERSION

    def _set_rp_delete_url(self, bookmark_model):
        return self.BOOKMARK_BASE_URL + bookmark_model.id + self.BOOKMARK_API_VERSION

    def _generate_bookmark_payload(self, bookmark_model):
        items = jsons.dump(bookmark_model, strip_privates=True)
        return jsons.dump(self._cleanup_json(items))

    def _cleanup_json(self, data):
        """
            Delete keys with the value ``None`` in a dictionary, recursively.
            This alters the input so you may wish to ``copy`` the dict first.
        """
        for key, value in list(data.items()):
            if key == 'bookmark_resource_base':
                del data[key]
            elif value is None:
                del data[key]
            elif isinstance(value, dict):
                self._cleanup_json(value)

        return data

    def add_bookmark(self, bookmark_model):
        """ Create a hunting bookmark """
        try:
            result = requests.put(
                self._set_rp_put_url(bookmark_model),
                headers=self._set_header(),
                json=self._generate_bookmark_payload(bookmark_model))
            print('Success')
            return result
        except Exception as e:
            print(str(e))

    def get_bookmarks(self, bookmark_model):
        """ Retrieve hunting bookmarks for workspace """
        try:
            result = requests.get(
                self._set_rp_get_url(bookmark_model),
                headers=self._set_header())
            print('Success')
            return result
        except Exception as e:
            print(str(e))

    def delete_bookmark(self, bookmark_model):
        """ Delete a hunting bookmark, not recommend for notebook users """
        try:
            result = requests.delete(
                self._set_rp_delete_url(bookmark_model),
                headers=self._set_header())
            print('Success')
            return result
        except Exception as e:
            print(str(e))

# end of the class



# pylint: disable-msg=C0103
# pylint: disable=line-too-long
"""
SentinelPortal: This package is developed for Sentinel notebooks integrating with Sentinel portal
"""

# __init__.py
from .bookmark_helper import Constants, BookmarkProperties, BookmarkModel, BookmarkHelper



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Anomaly Finder module:
This module has two classes: AnomalyQueries and AnomalyFinder
"""

import copy
import datetime as dt
import pandas as pd
from pandas.io.json import json_normalize
from azure.loganalytics.models import QueryBody

from SentinelUtils.obfuscation_utility import ObfuscationUtility
from SentinelLog.log import Log
from .anomaly_lookup_view_helper import AnomalyLookupViewHelper


class AnomalyQueries(): # pylint: disable=too-few-public-methods
    """ KQLs for anomaly lookup """

    KEY = b'7jH91qslpgxNSdVPiQwtr2D4bSg2i5ArfJqAMA1zCkU='
    QUERIES = {}
    # pylint: disable=line-too-long
    QUERIES['LISTTABLES'] = b'gAAAAABfdL95MQF9HV5YIbGGKy9kWPKXpbH3VWf80l9elru76D6sw8JN-shOQSYcWfSW43PfW7LSoShP14zl8_iu1-RC-mFYX_ZxiUBTZ5H-MHZZIAU0-nL1QOJ3EgJqHWpU79Kf9VsxjuKDf6oedEDnveb5MaM2Li-qQeJ7edUb1JknuQD5JFKRpqr_yE1jBQO3RPNxyXg9rhFPkNsTTxyHXZLE3NWQMA=='
    # pylint: disable=line-too-long
    QUERIES['ISCATCOLUMN'] = b'gAAAAABdNjkBT7XTAvPjCp56dI7LRL35EPMzF8UX_c4Hs0q910bTI2y-6viJGG7ZlXjclORrXFK68dAd2dYfLB0A_rw0Q9CeLEGUuWkHSFc7al2xwu7uEgliddsQQyocWhqyWlEtFEReIeJaqZYaSOkLon7sPN_icKEwiHc096kCkyjw5D0TeZ0Kgmnc5799Al7ND0kAk7KO'
    # pylint: disable=line-too-long
    QUERIES['ISCATHEURISTIC'] = b'gAAAAABdNjkglsyzKMCUkIXq3aqcim0F70S86HfAqaiyNUIF0La2st1DkiQTFK_vqVIyxiY25i78FiT6y0yZG4YQmpAVVwRJ302KkeAPVq0mPPK1FKbRcnnqmIc1HOAycyv3dmDHDUG7-_b-wyy8IDneWYyTE3TxyLUcG3kJRTmQd_6-hLXfDUjctjm0VPrA9zcrN8Il1y-nq-4jOsFZbO0qvHZfnLBTwaf52hkXPqmkZN9Rz-deW6Q4VY_j7Vw7rRrXM5WxRchL2kJBhGOq-hM8A3W9AA7qHnkgVu2BOVEYZAT_EnHvks8YMcWP04WKs49Dw5Ga4UMqJeU4MJH8PKfNmn7zcuLwMA=='
    # pylint: disable=line-too-long
    QUERIES['TIMESERIESANOMALYDETECTION'] = b'gAAAAABeNN8ojnOJUuvi1_FI8bqz200wGX1CPFnxU8FHDnwmFx7Ywm54WwbFXMWXVIpiLf9zjv5Fcl85wdyBdA6KS7_xph89Geb5CQOoqMGt_-syZ_KgE4CXQoCDCfWmCYXlx2zzIZX5g88SfVJeLmWbCUsk20KTO0Ecdt6TauIUuLkCh85v55_j8RSxXvwZy4y-WAate9hZh1xoRV5fjvq1_ox2V_qpJO-HpzdCuZGelMx3DkxL0PUy2_SXLZdaa5HLl0vM0IfmiDqIQgcxnjAEwk6GGezh3FvT43BCwO_HX5PHixCkqozeXzbsLuWxOMpoZ5I1174dVWK-8e9uiVvMfwyOqeC9tlKoAUXAEBsXS41kEbAMTwlK3VcCqq9iTfnW0jUkSlMV7P_JbWOVxErwC4DtIKP5d6AyNcqVhBlOYZwouqoY734vBhDsx95bzN2mwwL2-_Set_ksdoxTLlFVaWK9MY4tmBCsoQSVHrVuo00K1zPYIh43VgbVm6gFLs7kdjFDraMb5fjx0VxXQrkUkeO3xZ7i9caKl2ODpoVf2ahAPxqPTmHIG9bqQBb2SGu3PakxrRiRanOLY-pJ4eN54fQlYhk-O-1UwpUnna4WVMbqST4tFbEBCpr4IeGNqnaEeNccuHaWHDdjxIEt5sZ402LFJAV0HvMOxOVD3qUjUroHFysJrZXzYJF9HApBbtryokMjgh1YPGAwguniTTyUQolrV_1m54yhJLd7I-39MnBpfobO6sXAtZjPxZfoCCmiJta_JaAuj2sW-8goxU0NBWUxBsiEnPIrHyJKLiduavgID2UcHR4rhByyUg3EzbeKFO-WS-oIVXn0jFUfGjhOKc70Xc1R3L_vvGe2029AYkWmHb36yEJgVlVJ4YQ_7eMSjosZA6R97SDzYXCC1eOopH1hyexwvJwmlzP2gb3NcZ41WbmCTA7vYlT1uH4IqMiuvNWwkZ7MpPIccOnwFWkLFzh_BpgDfvnGy-loqaTNuuLUYzoAyhejzfBQPwG7AlPwH85pfNGarrz7z47uMYQE51-R2gxpDSmw3QvWacKIs3F2g94umQjXJkZ9otiNxPqZceKINig3pFj_SMRiQ1vXRNyKUM8BYptJJ7CFR5TI37lElejgoQBD3VQ7uyK_Ghz8M4cJG1P3ry9d8mk0-vwNClGkb4WGgLR69dPBSZM03uQ14oolEHxuwPQgBoLM-Rlu3YuGQBfbCtqtarLa8IwFBQcbA6WsKJ2dDnCcrRdCuMMTsKuJlSfoU-7jAOwVN-ISg8m47aWbJZtjAODzPxuv9KOQlTfApJcjla37UKKctce9kVkYeoqi6dJnSF9HLsfbRhSIf9bWKHue2ML20urH-0xhIDxvSA=='
    # pylint: disable=line-too-long
    QUERIES['TIMEWINDOWQUERY'] = b'gAAAAABdOjxI1Cq7frn_5Gj1l2vvA6Eu-a5qghqvRTBc8I9gWdcI8JiALXjpT7qJwf8ZBCKCrwYtMXY2-bp7Cj4jwYVXVDKmXRjoyz0xLbiVdCkIc07U2sNjpwzO1y1OvRr2apYv5Y9_yh_vOpqN4uv1WUezH_z1bXNCO-yI-LMIlidav4Xh5KwRtGBTnXGBk5YidPJVHfnZZGpCQ5w7g4t0ptoM5p6w_eXC8RZ82J3QLIVGtguWISFYweE5GWVJkkUXq3aq3n36uIFl2T3YllLUX2FytfOw_B8Xt1UrspWURfgDx1xqyCnqUEPG_EnO-TuGKFbMkh6AjpcduidHTuuS45YGatPvzRzyAzElLnbj7-s0gc-0POrUyiNaeTj_Tg0wTBsIJklL'
    # pylint: disable=line-too-long
    QUERIES['ISENTITYINTABLE'] = b'gAAAAABdNkO8YYV6ElbBqI9qp0oLHLquoYJD_7umEu1sDgyHouYcN0jU6vlOPp8AN5lecaMvXPUqQ5ZiFw6393Z9l7kNOB7IMITURv59MZJxeEVpt5ud9F4ge-5JGge5k7ux2YU50z-u9djJYet2SO-n1MpD5xO14ODKtBPsr9guZ40wYJwMzwLCjDSpTXFnIDjYrXDhfU3D2YGc4jnrq2EePBUAPPKxnIXg7AtmnGm4Add1_aV-pDlHMXTn09Z3kvlUcHpHBw7g'

    @staticmethod
    def get_query(name):
        """ get KQL """

        en_query = AnomalyQueries.QUERIES[name]
        obfuscate = ObfuscationUtility(AnomalyQueries.KEY)
        query = obfuscate.deobfuscate_text(en_query)
        return query


class AnomalyFinder():
    """
    This class provides process flow functions for anomaly lookup.
    Method - run is the main entry point.
    """

    def __init__(self, workspace_id, la_data_client):
        self.workspace_id = workspace_id
        self.la_data_client = la_data_client
        self.logger = Log()
        self.anomaly = ''

    def query_table_list(self):
        """ Get a list of data tables from Log Analytics for the user """

        query = AnomalyQueries.get_query('LISTTABLES')
        return self.query_loganalytics(query)

    def query_loganalytics(self, query):
        """ This method will call Log Analytics through LA client """

        res = self.la_data_client.query(self.workspace_id, QueryBody(query=query))
        json = res.as_dict()
        cols = json_normalize(json['tables'][0], 'columns')
        data_frame = json_normalize(json['tables'][0], 'rows')
        if data_frame.shape[0] != 0:
            data_frame.columns = cols.name
        return data_frame

    @staticmethod
    def construct_related_queries(df_anomalies):
        """ This method constructs query for user to repo and can be saves for future references """

        if df_anomalies.shape[0] == 0:
            return None

        queries = ''
        for tbl in df_anomalies.Table.unique():

            cur_table_anomalies = df_anomalies.loc[df_anomalies.Table == tbl, :]
            query = """{tbl} \
            | where TimeGenerated > datetime({maxTimestamp})-14d and TimeGenerated < datetime({maxTimestamp}) \
            | where {entCol} has "{qEntity}" \
            | where """.format(**{
                'tbl': tbl,
                'qTimestamp': cur_table_anomalies.qTimestamp.iloc[0],
                'maxTimestamp': cur_table_anomalies.maxTimestamp.iloc[0],
                'entCol': cur_table_anomalies.entCol.iloc[0],
                'qEntity': cur_table_anomalies.qEntity.iloc[0]
            })

            for j, row in cur_table_anomalies.iterrows(): # pylint: disable=unused-variable
                query += " {col} == to{colType}(\"{colVal}\") or".format(
                    col=row.colName,
                    colType=(row.colType) if 'colType' in row.keys() else 'string',
                    colVal=row.colVal.replace('"', '')
                )

            query = query[:-2] # drop the last or
            query += " | take 1000; " # limit the output size
            query = query.replace("\\", "\\\\")

            queries += query
        return queries

    # pylint: disable=too-many-locals
    def get_timewindow(self, q_entity, q_timestamp, ent_col, tbl):
        """ find the relevant time window for analysis """

        win_start = 0
        min_timestamp = None
        delta = None
        max_timestamp = None
        long_min_timestamp = None
        time_window_query_template = AnomalyQueries.get_query('TIMEWINDOWQUERY')

        for from_hour in range(-30, 0, 1):
            kql_time_range_d = time_window_query_template.format(
                table=tbl,
                qDate=q_timestamp,
                entColumn=ent_col,
                qEntity=q_entity,
                f=from_hour,
                t=from_hour+1,
                delta='d')

            df_time_range = self.query_loganalytics(kql_time_range_d)

            if df_time_range.shape[0] > 0:
                win_start = from_hour
                break

        dt_q_timestamp = pd.to_datetime(q_timestamp)
        ind2now = dt.datetime.utcnow() - dt_q_timestamp
        if win_start < -3:
            if ind2now > dt.timedelta(days=1):
                delta = '1d'
                max_timestamp = dt_q_timestamp + dt.timedelta(days=1)
            else:
                delta = '1d'
                max_timestamp = dt.datetime.now()
            long_min_timestamp = max_timestamp + dt.timedelta(days=win_start)
            min_timestamp = max_timestamp + dt.timedelta(days=max([-6, win_start]))

        elif win_start < 0: # switch to hours
            win_start_hour = -5
            for from_hour in range(-3*24, -5, 1):
                kql_time_range_h = time_window_query_template.format(
                    table=tbl,
                    qDate=q_timestamp,
                    entColumn=ent_col,
                    qEntity=q_entity,
                    f=from_hour,
                    t=from_hour+1,
                    delta='h')

                df_time_range = self.query_loganalytics(kql_time_range_h)

                if df_time_range.shape[0] > 0:
                    win_start_hour = from_hour
                    break
            if win_start_hour < -5:
                if ind2now > dt.timedelta(hours=1):
                    delta = '1h'
                    max_timestamp = dt_q_timestamp + dt.timedelta(hours=1)
                else:
                    delta = '1h'
                    max_timestamp = dt.datetime.now()
                min_timestamp = max_timestamp + dt.timedelta(hours=win_start_hour)
                long_min_timestamp = min_timestamp

        return min_timestamp, delta, max_timestamp, long_min_timestamp

    # pylint: disable=too-many-locals
    def run(self, q_timestamp, q_entity, tables):
        """ Main function for Anomaly Lookup """

        progress_bar = AnomalyLookupViewHelper.define_int_progress_bar()
        display(progress_bar)  # pylint: disable=undefined-variable

        # list tables if not given
        if not tables:
            kql_list_tables = AnomalyQueries.get_query('LISTTABLES')
            tables = self.query_loganalytics(kql_list_tables)
            tables = tables.SentinelTableName.tolist()

        progress_bar.value += 1

        # find the column in which the query entity appears in each table
        # - assumption that it appears in just one columns
        tables2search = []
        is_entity_in_table_template = AnomalyQueries.get_query('ISENTITYINTABLE')

        for tbl in tables:
            kql_entity_in_table = is_entity_in_table_template.format(
                table=tbl,
                qDate=q_timestamp,
                qEntity=q_entity)
            ent_in_table = self.query_loganalytics(kql_entity_in_table)

            if ent_in_table.shape[0] > 0:
                ent_col = [col for col in ent_in_table.select_dtypes('object').columns[1:] if
                           ent_in_table.loc[0, col] is not None
                           and ent_in_table.loc[:, col].str.contains(q_entity, case=False).all()]
                if ent_col:
                    ent_col = ent_col[0]
                tables2search.append({'table': tbl, 'entCol': ent_col})

        progress_bar.value += 2

        # for each table, find the time window to query on
        for tbl in tables2search:
            tbl['minTimestamp'], tbl['delta'], tbl['maxTimestamp'], tbl['longMinTimestamp'] = \
            self.get_timewindow(q_entity, q_timestamp, tbl['entCol'], tbl['table'])

        progress_bar.value += 1

        # identify all the categorical columns per table on which we will find anomalies
        categorical_cols = []
        is_cat_column_template = AnomalyQueries.get_query('ISCATCOLUMN')
        is_cat_heuristic_template = AnomalyQueries.get_query('ISCATHEURISTIC')
        for tbl in tables2search:
            kql_is_cat_column = is_cat_column_template.format(table=tbl['table'])
            df_cols = self.query_loganalytics(kql_is_cat_column)

            for col in df_cols.ColumnName:
                kql_is_cat_heuristic = is_cat_heuristic_template.format(
                    table=tbl['table'],
                    column=col)
                df_is_cat = self.query_loganalytics(kql_is_cat_heuristic)

                if df_is_cat.shape[0] > 0:
                    cat_col_info = copy.deepcopy(tbl)
                    cat_col_info['col'] = col
                    categorical_cols.append(cat_col_info)

        progress_bar.value += 2

        anomalies_list = []
        time_series_anomaly_detection_template = \
            AnomalyQueries.get_query('TIMESERIESANOMALYDETECTION')
        for col_info in categorical_cols:
            max_timestamp = col_info['maxTimestamp'].strftime('%Y-%m-%dT%H:%M:%S.%f')
            long_min_timestamp = col_info['longMinTimestamp'].strftime('%Y-%m-%dT%H:%M:%S.%f')

            kql_time_series_anomaly_detection = time_series_anomaly_detection_template.format(
                table=col_info['table'],
                column=col_info['col'],
                entColumn=col_info['entCol'],
                qEntity=q_entity,
                minTimestamp=long_min_timestamp,
                maxTimestamp=max_timestamp,
                qTimestamp=q_timestamp,
                delta=col_info['delta'])

            cur_anomalies = self.query_loganalytics(kql_time_series_anomaly_detection)

            anomalies_list.append(cur_anomalies)

        progress_bar.value += 2

        if anomalies_list:
            anomalies = pd.concat(anomalies_list, axis=0)
        else:
            anomalies = pd.DataFrame()

        progress_bar.value += 2
        queries = AnomalyFinder.construct_related_queries(anomalies)
        progress_bar.close()
        self.anomaly = str(anomalies.to_json(orient='records'))

        return anomalies, queries
# End of the Module #



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Version Management:
This module is used to validate installed Python packages
that are required by Azure Sentinel Notebooks.
"""

import sys
import pkg_resources

# pylint: disable-msg=R0903
class VersionInformation:
    """ object model for version information """
    name = ''
    current_version = ''
    required_version = ''
    requirement_met = False
    message = ''

# pylint: disable-msg=R0201
class ModuleVersionCheck:
    """ class implements version checking """

    # pylint: disable=line-too-long
    def validate_python(self, required_version):
        """ validating Python version """
        version = VersionInformation()
        version.name = 'Python'
        version.current_version = sys.version
        version.required_version = required_version
        version.requirement_met = sys.version_info >= tuple(int(x) for x in required_version.split("."))
        version.message = VersionInformation.name + required_version + ' is required' if not VersionInformation.requirement_met else ''
        return version

    # pylint: disable-msg=W0702
    def validate_installed_modules(self, module_list):
        """ validating installed modules' version """
        module_versions = []
        for mod_info in module_list:
            version = VersionInformation()
            version.name, version.required_version = mod_info.split(">=")
            try:
                pkg_resources.require(mod_info)
                version.requirement_met = True
            except:
                version.requirement_met = False
                try:
                    version.message = str(sys.exc_info()[0].report)
                except:
                    version.message = 'Unknown error'
            finally:
                if 'azure' in version.name and version.message.find("VersionConflict") >= 0:
                    version.requirement_met = True
                elif version.message == '' or version.message.find("DistributionNotFound") < 0:
                    version = self.get_version_information(version, mod_info)

            module_versions.append(version)
        return module_versions

    # pylint: disable-msg=W0613
    def get_version_information(self, version, mod_info):
        """ get version information """
        mod_list = pkg_resources.WorkingSet()
        items = list(filter(lambda x: x.project_name.startswith(version.name), mod_list))
        if not items:
            version.current_version = items[0].version
        return version

# end of the class



# -------------------------------------------------------------------------

# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
input_error module:
This module include custom exceptions
"""


class InputError(ValueError):

    """ This class provides custom exception for validating input data """



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
obfuscation_utility module:
This module provides obfuscation functionalities
"""

from cryptography.fernet import Fernet


class ObfuscationUtility():
    """ This class provides utility methods for obfuscation """
    def __init__(self, seed):
        self.seed = seed

    @staticmethod
    def generate_seed():
        """ Generate seed """

        return Fernet.generate_key()

    def obfuscate_text(self, text):
        """ Obfuscate input text using key """

        fernet = Fernet(self.seed)
        en_text = fernet.encrypt(text)
        return en_text

    def deobfuscate_text(self, en_text):
        """ De-obfuscate input text using key """

        fernet = Fernet(self.seed)
        re_text = fernet.decrypt(en_text)
        return re_text.decode('utf-8')



"""
 Copyright (c) Microsoft Corporation. All rights reserved.
 Licensed under the MIT License. See License.txt in the project root for
 license information.
"""

import setuptools

INSTALL_REQUIRES = ['azure-loganalytics>=0.1.0', 
                    'azure-mgmt-loganalytics>=0.2.0',
                    'azure-common>=1.1.25',
                    'azure-cli-core>=2.7.0'
                    ]

with open("LICENSE.txt", "r") as fh:
    LICENSE_TXT = fh.read()

setuptools.setup(
    name="Azure-Sentinel-Utilities",
    version="0.6.5",
    author="Azure Sentinel Notebooks Devs",
    author_email="zhzhao@microsoft.com",
    description="AZURE SENTINEL NOTEBOOKS PYTHON TOOLS: \
    This package is developed to support Azure Sentinel Notebooks.  \
    It is in an early preview stage so please provide feedback, \
    report bugs, and suggets for new features.",
    license=LICENSE_TXT,
    url="https://github.com/Azure/Azure-Sentinel",
    python_requires='>=3.6',
    packages=setuptools.find_packages(),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
    ],
    install_requires=INSTALL_REQUIRES,
    keywords=['security', 'azure', 'sentinel'],
    zip_safe=False,
)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Widget View Helper:
This module provides helper methods for UI controls and components.
"""

import os
import ipywidgets as widgets
from IPython.display import HTML

# pylint: disable-msg=R0904
# pylint: disable-msg=E0602
class WidgetViewHelper():
    """ This classes provides helper methods for UI controls and components. """

    def __init__(self):
        self.variable = None

    def set_env(self, env_dir, var):
        """ Set notebook environmental variable """
        val = None
        if var in env_dir.keys():
            val = env_dir[var]

        self.variable = var
        user_input = widgets.Text(value=val, description=var + ': ')
        user_input.observe(self.save_env, 'value')
        display(user_input)

    def save_env(self, val):
        """ Save notebook environmental variable """
        os.environ[self.variable] = val.new

    @staticmethod
    def set_env1(reset, env_dir, env_dict):
        """ Set notebook environmental variables """
        for key in env_dict.keys():
            user_input = ''
            if not reset and key in env_dir:
                env_dict[key] = env_dir[key]
                print(key + '=' + env_dir[key])
            else:
                user_input = widgets.Text(description=key + ': ')
                os.environ[key] = user_input.value
                env_dict[key] = user_input.value
        return env_dict

    @staticmethod
    def select_vm(compute):
        """ Select a VM """
        vm_names = sorted(list(vm.name for vm in list(compute.get_vm_list())))
        return widgets.Dropdown(options=vm_names, value=vm_names[0], description='VM:')

    @staticmethod
    def select_managed_disk(compute, vm_name):
        """ Select a managed disk """
        disk_list = compute.get_vm_disk_names(vm_name)
        return widgets.Dropdown(options=disk_list, value=disk_list[0], description='Disk:')

    @staticmethod
    def select_account_creation():
        """ Create a new account or use existing account """
        storage_account_creation = ['Creating new account', 'Using exist account']
        return widgets.Dropdown(options=storage_account_creation,
                                value=storage_account_creation[0],
                                description='Storage Account Creation:')

    @staticmethod
    def select_blob_container_creation():
        """ Create a new container or use existing container """
        blob_container_creation = ['Creating new container', 'Using exist container']
        return widgets.Dropdown(options=blob_container_creation,
                                value=blob_container_creation[0],
                                description='Blob Container Creation:')

    @staticmethod
    def select_os():
        """ Select Windows or Linux """
        os_type_list = ['Windows', 'Linux']
        return widgets.Dropdown(options=os_type_list, value=os_type_list[0], description='OS Type:')

    @staticmethod
    def check_storage_account_name_availability(storage):
        """ Check if a storage account name is available to use """
        # usert input storage account name
        storage_account_name = input('Storage Account Name:')
        name_availability = storage.is_storage_account_name_available(storage_account_name)
        return storage_account_name if name_availability.name_available else None

    @staticmethod
    def create_storage_account_and_get_key(storage,
                                           storage_account_name,
                                           resource_group_for_storage):
        """ Create storage account """
        storage_location = input('Storage Location:')
        storage.create_storage_account_async(
            storage_account_name,
            resource_group_for_storage,
            **{'storage_location' : storage_location})
        return storage.get_storage_account_key(storage_account_name, resource_group_for_storage)

    @staticmethod
    def select_storage_account(storage, resource_group_for_storage):
        """ Select a storage account """
        storage_account_list = storage.get_storage_account_names(resource_group_for_storage)
        return widgets.Dropdown(options=storage_account_list,
                                value=storage_account_list[0],
                                description='Existing Storage Accounts:')

    @staticmethod
    def select_blob_container(storage, resource_group_for_storage, storage_account_name):
        """ Select a blob container """
        blob_container_list = storage.get_container_name_list(resource_group_for_storage,
                                                              storage_account_name,
                                                              None)
        return widgets.Dropdown(options=blob_container_list,
                                value=blob_container_list[0],
                                description='Blob Containers:')

    @staticmethod
    def select_log_analytics_workspace(loganalytics):
        """ Select a LA workspace """
        workspace_name_list = loganalytics.get_workspace_name_list()
        return widgets.Dropdown(options=workspace_name_list,
                                value=workspace_name_list[0],
                                description='Workspace:')

    @staticmethod
    def select_multiple_tables(anomaly_lookup):
        """ Select data tables """
        table_list = anomaly_lookup.query_table_list()
        tables = sorted(table_list.SentinelTableName.tolist())
        return widgets.SelectMultiple(options=tables,
                                      row=len(tables),
                                      value=[],
                                      description='Tables:')

    @staticmethod
    def generate_upload_container_path(storage, os_type, sas_expiration_in_days):
        """ Generate a upload container path """
        sas_url = storage.generate_blob_container_sas_url(sas_expiration_in_days)
        upload_container_path = storage.build_upload_container_path(os_type, sas_url)
        return upload_container_path

    @staticmethod
    # pylint: disable=line-too-long
    def get_vm_extension_properties(os_type, upload_container_path, user_id=None):
        """ Get VM extensions properties """
        if os_type == 'Windows':
            command_to_execute = 'powershell -File installNotebookExtension.ps1 "{0}" >> out.txt'.format(upload_container_path)
            file_list = ['https://sentinelnotebooks.blob.core.windows.net/piwindowsstorage/installNotebookExtension.ps1',
                         'https://sentinelnotebooks.blob.core.windows.net/piwindowsstorage/piextension.zip']
        elif os_type == 'Linux':
            command_to_execute = './piondemand.sh "' + upload_container_path + '"'
            file_list = ['https://sentinelnotebooks.blob.core.windows.net/pilinuxstorage/release/ondemand/stable/piondemand.sh',
                         'https://sentinelnotebooks.blob.core.windows.net/pilinuxstorage/release/ondemand/stable/pilinux.ondemand.tar.bz2']

        elif os_type == 'DSVM':
            command_to_execute = './azureforensics.sh {0}'.format(user_id)
            file_list = ['https://sentinelnotebooks.blob.core.windows.net/forensicsnotebooks/azureforensics.sh',
                         'https://sentinelnotebooks.blob.core.windows.net/forensicsnotebooks/vhdexplorer.tar']

        return command_to_execute, file_list

    @staticmethod
    def define_int_progress_bar():
        """ Define a progress bar """
        return widgets.IntProgress(value=0,
                                   min=0,
                                   max=10,
                                   step=1,
                                   description='Loading:',
                                   bar_style='success',
                                   orientation='horizontal',
                                   position='top')

    @staticmethod
    # pylint: disable=line-too-long
    def copy_to_clipboard(url, text_body, label_text):
        """ Copy text to Clipboard """
        html_str = (
            """<!DOCTYPE html>
            <html><body style="height:20px">
            <input  id="sentinel_text_for_copy" type="text" readonly style="font-weight: bold; border: none; max-height:10px; width:1px;" size = '"""
            + str(len(text_body))
            + """' value='"""
            + text_body
            + """'>
            <button style="border: 2px solid #4CAF50;" onclick="sentinel_copy()">""" + label_text + """</button>
            <script>
            function sentinel_copy() {
                var copyText = document.getElementById("sentinel_text_for_copy");
                copyText.select();
                document.execCommand("copy");
            }
            </script>
            </body></html>"""
        )

        return html_str

    @staticmethod
    # pylint: disable=line-too-long
    def construct_url_for_log_analytics_logs(tenant_domain,
                                             subscription_id,
                                             resource_group,
                                             workspace_name):
        """ Generate URL for LA logs """
        return 'https://portal.azure.com/#blade/Microsoft_Azure_Security_Insights/MainMenuBlade/7/subscriptionId/{0}/resourceGroup/{1}/workspaceName/{2}'.format(subscription_id, resource_group, workspace_name)

    @staticmethod
    # pylint: disable=undefined-variable
    def display_html(inner_html):
        """ Display HTML """
        display(HTML(inner_html))

    @staticmethod
    def pick_start_and_end_date():
        """ Pick dates """
        start_date = widgets.DatePicker(description='Pick a start date', disabled=False)
        end_date = widgets.DatePicker(description='Pick a end date', disabled=False)
        # pylint: disable=undefined-variable
        display(start_date)
        # pylint: disable=undefined-variable
        display(end_date)
        return start_date, end_date

    @staticmethod
    # pylint: disable=line-too-long
    # pylint: disable=undefined-variable
    def select_multiple_items(label, item_name):
        """ Select multiple items """
        label_item = widgets.Label(value=label)
        items = widgets.Textarea(value='', placeholder='One per line: \n 0x7ae3 \n 0x7ae6', description=item_name, disabled=False, rows=5)
        display(label_item)
        display(items)
        return items
